@@ -6,7 +6,7 @@
 /*                                                                          */
 /*                          C Implementation File                           */
 /*                                                                          */
-/*                            $Revision: 1.88 $                             */
+/*                            $Revision: 1.110 $                             */
 /*                                                                          */
 /*           Copyright (c) 1992,1993,1994 NYU, All Rights Reserved          */
 /*                                                                          */
@@ -46,6 +46,9 @@
 #undef  NULL
 #define NULL 0
 
+/* If nonzero, pretend we are allocating at global level.  */
+
+int force_global;
 
 /* Global Variables for the various types we create.  */ 
 
@@ -60,7 +63,9 @@ tree ptr_void_type_node;
 tree void_ftype;
 tree ptr_void_ftype;
 
+tree except_type_node;
 tree malloc_decl;
+tree free_decl;
 tree jmpbuf_type;
 tree jmpbuf_ptr_type;
 tree get_jmpbuf_decl;
@@ -68,7 +73,6 @@ tree set_jmpbuf_decl;
 tree get_excptr_decl;
 tree raise_decl;
 tree raise_nodefer_decl;
-tree reraise_decl;
 tree setjmp_decl;
 tree raise_constraint_error_decl;
 tree unchecked_union_node;
@@ -196,7 +200,8 @@ static struct binding_level clear_binding_level = {NULL, NULL, NULL, NULL};
 int
 global_bindings_p ()
 {
-  return current_binding_level == global_binding_level ? -1 : 0;
+  return (force_global != 0 || current_binding_level == global_binding_level
+	  ? -1 : 0);
 }
 
 /* Return the list of declarations in the current level. Note that this list
@@ -299,12 +304,12 @@ poplevel (keep, reverse, functionbody)
     {
       block_node = block_created_by_back_end;
 
-      /* Check if we are about to discard some information that was gathered
-	 by the front-end. Nameley check if the back-end created a new block 
-	 without calling pushlevel first. To understand why things are lost
-	 just look at the next case (i.e. no block created by back-end.  */
+      /* Update decls and chain into the block the back end made.  */
       if ((keep || functionbody) && (decl_chain || subblock_chain))
-	gigi_abort (403);
+	{
+	  BLOCK_VARS (block_node) = keep? decl_chain : 0;
+	  BLOCK_SUBBLOCKS (block_node) = subblock_chain;
+	}
     }
 
   /* If there were any declarations in the current binding level, or if this
@@ -427,9 +432,11 @@ pushdecl (decl)
      types, so let's not try it (at least for now).  */
 
   if (TREE_CODE (decl) == TYPE_DECL
+      && DECL_NAME (decl) != 0
       && (TYPE_NAME (TREE_TYPE (decl)) == 0
-	  || *(IDENTIFIER_POINTER (TYPE_NAME (TREE_TYPE (decl)))) == 'T'))
-    TYPE_NAME (TREE_TYPE (decl)) = DECL_NAME (decl);
+	  || (TREE_CODE (TYPE_NAME (TREE_TYPE (decl))) == IDENTIFIER_NODE
+	      && *(IDENTIFIER_POINTER (TYPE_NAME (TREE_TYPE (decl)))) == 'T')))
+    TYPE_NAME (TREE_TYPE (decl)) = decl;
 
   return decl;
 }
@@ -515,12 +522,22 @@ init_decl_processing ()
 
   /* malloc is a function declaration tree for a function to allocate
      memory.  */
-  malloc_decl = create_subprog_decl ("malloc", NULL_PTR,
+  malloc_decl = create_subprog_decl ("__gnat_malloc", NULL_PTR,
 				     build_function_type (ptr_void_type_node,
 							  tree_cons (NULL_TREE,
 								     sizetype,
 								     endlink)),
-				     NULL_TREE, 0, 1, 1, 0);
+				     NULL_TREE, 0, 1, 1, 0, NULL_TREE);
+
+  /* free is a function declaration tree for a function to free memory.  */
+
+  free_decl
+    = create_subprog_decl ("free", NULL_PTR,
+			   build_function_type (void_type_node,
+						tree_cons (NULL_TREE,
+							   ptr_void_type_node,
+							   endlink)),
+			   NULL_TREE, 0, 1, 1, 0, NULL_TREE);
 
   /* Make the types and functions used for exception processing.  
      We assume here that a jmp_buf is no more than 100 size_t entries.  */
@@ -534,7 +551,7 @@ init_decl_processing ()
     = create_subprog_decl ("system__task_specific_data__get_jmpbuf_address",
 			   NULL_PTR,
 			   (build_function_type (jmpbuf_ptr_type, NULL_TREE)),
-			   NULL_TREE, 0, 1, 1, 0);
+			   NULL_TREE, 0, 1, 1, 0, NULL_TREE);
 
   set_jmpbuf_decl
     = create_subprog_decl ("system__task_specific_data__set_jmpbuf_address",
@@ -543,64 +560,62 @@ init_decl_processing ()
 					       tree_cons (NULL_TREE,
 							  jmpbuf_ptr_type,
 							  endlink)),
-			   NULL_TREE, 0, 1, 1, 0);
+			   NULL_TREE, 0, 1, 1, 0, NULL_TREE);
 
-  /* Function to get the current exception.  An exception is a single byte.  */
+  /* Right now the type of an exception is a byte.  We need to
+     get the actual type from the front end eventually.  */
+  except_type_node = char_type_node;
+
+  /* Function to get the current exception.  */
   get_excptr_decl
     = create_subprog_decl ("system__task_specific_data__get_gnat_exception",
 			   NULL_PTR,
 			   build_function_type
-			   (build_pointer_type (char_type_node), NULL_TREE),
-			   NULL_TREE, 0, 1, 1, 0);
+			   (build_pointer_type (except_type_node), NULL_TREE),
+			   NULL_TREE, 0, 1, 1, 0, NULL_TREE);
 
-  /* Functions that raise and re-raise exceptions.  */
+  /* Functions that raise exceptions.  */
   raise_decl
     = create_subprog_decl
       ("__gnat_raise", NULL_PTR,
        build_function_type (void_type_node,
 			    tree_cons (NULL_TREE,
-				       build_pointer_type (char_type_node),
+				       build_pointer_type (except_type_node),
 				       endlink)),
-       NULL_TREE, 0, 1, 1, 0);
+       NULL_TREE, 0, 1, 1, 0, NULL_TREE);
 
   raise_nodefer_decl
     = create_subprog_decl
       ("__gnat_raise_nodefer", NULL_PTR,
        build_function_type (void_type_node,
 			    tree_cons (NULL_TREE,
-				       build_pointer_type (char_type_node),
+				       build_pointer_type (except_type_node),
 				       endlink)),
-       NULL_TREE, 0, 1, 1, 0);
-
-  /* Operand is 1 if user call, else 0.  */
-  reraise_decl
-    = create_subprog_decl ("__gnat_reraise", NULL_PTR,
-			   build_function_type (void_type_node,
-						tree_cons (NULL_TREE,
-							   integer_type_node,
-							   endlink)),
-			   NULL_TREE, 0, 1, 1, 0);
+       NULL_TREE, 0, 1, 1, 0, NULL_TREE);
 
   /* Indicate that these never return.  */
   TREE_THIS_VOLATILE (raise_decl) = 1;
   TREE_SIDE_EFFECTS (raise_nodefer_decl) = 1;
-  TREE_SIDE_EFFECTS (reraise_decl) = 1;
 
   /* setjmp returns an integer and has one operand, which is a pointer to
      a jmpbuf.  */
   setjmp_decl
     = create_subprog_decl
+#ifdef WINNT
+      ("_setjmp", NULL_PTR,
+#else
       ("setjmp", NULL_PTR,
+#endif
        build_function_type (integer_type_node,
 			    tree_cons (NULL_TREE,  jmpbuf_ptr_type, endlink)),
-       NULL_TREE, 0, 1, 1, 0);
+       NULL_TREE, 0, 1, 1, 0, NULL_TREE);
 
   /* __gnat_raise_constraint takes no operands and never returns.  */
   raise_constraint_error_decl
     = create_subprog_decl
       ("__gnat_raise_constraint_error", NULL_PTR,
        build_function_type (void_type_node, endlink),
-       NULL_TREE, 0, 1, 1, 0);
+       NULL_TREE, 0, 1, 1, 0, NULL_TREE);
 
   TREE_THIS_VOLATILE (raise_constraint_error_decl) = 1;
   TREE_SIDE_EFFECTS (raise_constraint_error_decl) = 1;
@@ -634,17 +649,95 @@ finish_incomplete_decl (dont_care)
 }
 
 /* Given a record type (RECORD_TYPE) and a chain of FIELD_DECL
-   nodes (FIELDLIST), finish constructing the record or union type.  */
+   nodes (FIELDLIST), finish constructing the record or union type. 
+   If HAS_REP is nonzero, this record has a rep clause; don't call
+   layout_type but merely set the size and alignment ourselves.  */
 
 void
-finish_record_type (record_type, fieldlist)
+finish_record_type (record_type, fieldlist, has_rep)
      tree record_type;
      tree fieldlist;
+     int has_rep;
 {
   TYPE_FIELDS (record_type) = fieldlist;
-  TYPE_STUB_DECL (record_type)
-    = pushdecl (build_decl (TYPE_DECL, NULL_TREE, record_type));
-  layout_type (record_type);
+
+  if (TYPE_NAME (record_type) != 0
+      && TREE_CODE (TYPE_NAME (record_type)) == TYPE_DECL)
+    TYPE_STUB_DECL (record_type) = TYPE_NAME (record_type);
+  else
+    TYPE_STUB_DECL (record_type)
+      = pushdecl (build_decl (TYPE_DECL, TYPE_NAME (record_type),
+			      record_type));
+
+  /* If we had a rep clause, compute the size from the highest ending position
+     plus one and the alignment from the highest actual alignment.  Otherwise,
+     let GCC lay out the type.  */
+  if (has_rep)
+    {
+      tree field;
+      int must_be_blkmode = 0;
+
+      TYPE_ALIGN (record_type) = MAX (BITS_PER_UNIT, TYPE_ALIGN (record_type));
+      TYPE_MODE (record_type) = BLKmode;
+      if (TYPE_SIZE (record_type) == 0)
+	TYPE_SIZE (record_type) = size_int (0);
+
+      for (field = fieldlist; field; field = TREE_CHAIN (field))
+	{
+	  tree end_bit = size_binop (PLUS_EXPR, DECL_FIELD_BITPOS (field),
+				     DECL_SIZE (field));
+	  int bitpos;
+
+	  if (tree_int_cst_lt (TYPE_SIZE (record_type), end_bit))
+	    TYPE_SIZE (record_type) = end_bit;
+
+	  if (TREE_INT_CST_LOW (DECL_FIELD_BITPOS (field)) % DECL_ALIGN (field) == 0
+	      && DECL_FIELD_SIZE (field) % DECL_ALIGN (field) == 0)
+	    TYPE_ALIGN (record_type)
+	      = MAX (TYPE_ALIGN (record_type), DECL_ALIGN (field));
+
+	  /* A record which has any BLKmode members must itself be BLKmode; it
+	     can't go in a register unless the member is BLKmode only because
+	     it isn't aligned.  */
+	  if (TYPE_MODE (TREE_TYPE (field)) == BLKmode
+	      && ! TYPE_NO_FORCE_BLK (TREE_TYPE (field)))
+	    must_be_blkmode = 1;
+
+	  bitpos = TREE_INT_CST_LOW (DECL_FIELD_BITPOS (field));
+
+	  /* Must be BLKmode if any field crosses a word boundary,
+	     since extract_bit_field can't handle that in registers.  */
+	  if (bitpos / BITS_PER_WORD
+	      != ((TREE_INT_CST_LOW (DECL_SIZE (field)) + bitpos - 1)
+		  / BITS_PER_WORD)
+	      /* But there is no problem if the field is entire words.  */
+	      && TREE_INT_CST_LOW (DECL_SIZE (field)) % BITS_PER_WORD == 0)
+	    must_be_blkmode = 1;
+	}
+
+      if (! must_be_blkmode)
+	TYPE_MODE (record_type)
+	  = mode_for_size (TREE_INT_CST_LOW (TYPE_SIZE (record_type)),
+			   MODE_INT, 1);
+
+      /* If structure's known alignment is less than
+	 what the scalar mode would need, and it matters,
+	 then stick with BLKmode.  */
+      if (STRICT_ALIGNMENT
+	  && ! (TYPE_ALIGN (record_type) >= BIGGEST_ALIGNMENT
+		|| (TYPE_ALIGN (record_type)
+		    >= TREE_INT_CST_LOW (TYPE_SIZE (record_type)))))
+	{
+	  if (TYPE_MODE (record_type) != BLKmode)
+	    /* If this is the only reason this type is BLKmode,
+	       then don't force containing types to be BLKmode.  */
+	    TYPE_NO_FORCE_BLK (record_type) = 1;
+	  TYPE_MODE (record_type) = BLKmode;
+	}
+    }
+  else
+    layout_type (record_type);
+
   rest_of_type_compilation (record_type, global_bindings_p ());
 }
 
@@ -760,13 +853,17 @@ create_type_decl (type_name, type)
 {
   tree id_node   = type_name ? get_identifier (type_name): NULL_TREE;
   tree type_decl = build_decl (TYPE_DECL, id_node, type);
+  enum tree_code code = TREE_CODE (type);
 
   /* Add this decl to the current binding level.  */
   type_decl = pushdecl (type_decl);
 
   /* Pass type declaration information to the debugger unless this is an
-     UNCONSTRAINED_ARRAY_TYPE, which the debugger does not support.  */
-  if (TREE_CODE (type) != UNCONSTRAINED_ARRAY_TYPE)
+     UNCONSTRAINED_ARRAY_TYPE, which the debugger does not support,
+     and ENUMERAL_TYPE or RECORD_TYPE which is handled separately,
+     or a dummy type, which will be completed later.  */
+  if (code != UNCONSTRAINED_ARRAY_TYPE && code != ENUMERAL_TYPE 
+      && code != RECORD_TYPE && ! TYPE_IS_DUMMY_P (type))
     rest_of_decl_compilation (type_decl, NULL, global_bindings_p (), 0);
 
   return type_decl;
@@ -783,6 +880,9 @@ static tree pending_elaborations;
    its data type (a GCC ..._TYPE node).  VAR_INIT is the GCC tree for an
    optional initial expression; NULL_TREE if none.
 
+   SIZE, if nonzero, is a GCC tree to be used for the size of the variable.
+   ALIGN, if nonzero, is the required alignment of the variable.
+
    CONST_FLAG is nonzero if this variable is constant.
 
    PUBLIC_FLAG is nonzero if this definition is to be made visible outside of
@@ -794,12 +894,14 @@ static tree pending_elaborations;
    it indicates whether to always allocate storage to the variable.   */
 
 tree
-create_var_decl (var_name, asm_name, type, var_init, const_flag, public_flag,
-		 extern_flag, static_flag)
+create_var_decl (var_name, asm_name, type, var_init, size, align,
+		 const_flag, public_flag, extern_flag, static_flag)
      char *var_name;
      char *asm_name;
      tree type;
      tree var_init;
+     tree size;
+     int align;
      int const_flag;
      int public_flag;
      int extern_flag;
@@ -831,9 +933,7 @@ create_var_decl (var_name, asm_name, type, var_init, const_flag, public_flag,
       var_init = 0;
     }
 
-#ifdef DECL_COMMON
   DECL_COMMON   (var_decl) = 1;
-#endif
   DECL_INITIAL  (var_decl) = var_init;
   TREE_READONLY (var_decl) = const_flag;
   DECL_EXTERNAL (var_decl) = extern_flag;
@@ -851,12 +951,19 @@ create_var_decl (var_name, asm_name, type, var_init, const_flag, public_flag,
   if (asm_name)
     DECL_ASSEMBLER_NAME (var_decl) = get_identifier (asm_name);
 
+  if (size != 0)
+    DECL_SIZE (var_decl) = size;
+
   /* If the type is a RECORD_TYPE and its size depends on a discriminant,
      the size to be used for the object is the maximum possible size.  */
-  if (TREE_CODE (type) == RECORD_TYPE && ! TREE_CONSTANT (TYPE_SIZE (type))
+  else if (TREE_CODE (type) == RECORD_TYPE
+	   && ! TREE_CONSTANT (TYPE_SIZE (type))
       && contains_placeholder_p (TYPE_SIZE (type)))
     DECL_SIZE (var_decl) = max_size (TYPE_SIZE (type), 1);
 
+  if (align != 0)
+    DECL_ALIGN (var_decl) = align;
+
   /* Add this decl to the current binding level and generate any
      needed code and RTL. */
   var_decl = pushdecl (var_decl);
@@ -870,23 +977,46 @@ create_var_decl (var_name, asm_name, type, var_init, const_flag, public_flag,
 
 /* Returns a FIELD_DECL node. FIELD_NAME the field name, FIELD_TYPE is its
    type, and RECORD_TYPE is the type of the parent.  PACKED is nonzero if
-   this field is in a record type with a "pragma pack".  */
+   this field is in a record type with a "pragma pack".  If SIZE is nonnegative
+   it is the specified size for this field and POS is the bit position.  */
 
 tree
-create_field_decl (field_name, field_type, record_type, packed)
+create_field_decl (field_name, field_type, record_type, packed, size, pos)
      char *field_name;
      tree field_type;
      tree record_type;
      int packed;
+     int size, pos;
 {
   tree field_id   = field_name ? get_identifier (field_name): NULL_TREE;
   tree field_decl = build_decl (FIELD_DECL, field_id, field_type);
 
   DECL_CONTEXT (field_decl) = record_type;
 
+  /* If we have a specified size and it is not the same as the size
+     of the mode or if the position does not correspond to the
+     proper alignment, make this a bit field.  Likewise for a BLKmode
+     field not byte aligned.  */
+  if (size >= 0)
+    {
+      if (((INTEGRAL_TYPE_P (field_type)
+	    && (size != GET_MODE_BITSIZE (TYPE_MODE (field_type))
+		|| pos % GET_MODE_ALIGNMENT (TYPE_MODE (field_type) != 0)))
+	  || (TYPE_MODE (field_type) == BLKmode
+	      && pos % BITS_PER_UNIT != 0)))
+	{
+	  DECL_BIT_FIELD (field_decl) = 1;
+	  DECL_PACKED (field_decl) = 1;
+	}
+
+      DECL_FIELD_SIZE (field_decl) = size;
+      DECL_FIELD_BITPOS (field_decl) = size_int (pos);
+      layout_decl (field_decl, pos);
+    }
+
   /* If this is to be packed and it is an integral type make a bit field whose
      width is the precision of the type.  */
-  if (packed && INTEGRAL_TYPE_P (field_type)
+  else if (packed && INTEGRAL_TYPE_P (field_type)
       && TREE_CODE (TYPE_SIZE (field_type)) == INTEGER_CST)
     {
       DECL_BIT_FIELD (field_decl) = 1;
@@ -973,7 +1103,8 @@ create_label_decl (label_name)
 
 tree
 create_subprog_decl (subprog_name, asm_name, subprog_type, param_decl_list,
-		     inline_flag, public_flag, extern_flag, pure_flag)
+		     inline_flag, public_flag, extern_flag, pure_flag,
+		     machine_attr)
      char *subprog_name;
      char *asm_name;
      tree subprog_type;
@@ -982,18 +1113,29 @@ create_subprog_decl (subprog_name, asm_name, subprog_type, param_decl_list,
      int public_flag;
      int extern_flag;
      int pure_flag;
+     tree machine_attr;
 {
   tree subprog_id   = get_identifier (subprog_name);
   tree return_type  = TREE_TYPE (subprog_type);
   tree subprog_decl = build_decl (FUNCTION_DECL, subprog_id, subprog_type);
 
-  DECL_EXTERNAL (subprog_decl) = extern_flag;
-  TREE_PUBLIC (subprog_decl)   = public_flag;
-  DECL_INLINE (subprog_decl)   = inline_flag;
-  TREE_READONLY (subprog_decl) = pure_flag;
+  DECL_EXTERNAL (subprog_decl)  = extern_flag;
+  TREE_PUBLIC (subprog_decl)    = public_flag;
+  DECL_INLINE (subprog_decl)    = inline_flag;
+  TREE_READONLY (subprog_decl)  = pure_flag;
   DECL_ARGUMENTS (subprog_decl) = param_decl_list;
   DECL_RESULT (subprog_decl)    = build_decl (RESULT_DECL, 0, return_type);
 
+  if (machine_attr)
+#if defined (VALID_MACHINE_DECL_ATTRIBUTE) \
+    || defined (VALID_MACHINE_TYPE_ATTRIBUTE)
+    if (valid_machine_attribute (machine_attr, subprog_decl, subprog_type))
+      ;
+    else
+#endif
+      warning ("`%s' attribute directive ignored",
+               IDENTIFIER_POINTER (TREE_PURPOSE (machine_attr)));
+
   if (asm_name)
     DECL_ASSEMBLER_NAME (subprog_decl) = get_identifier (asm_name);
 
@@ -1223,7 +1365,21 @@ tree
 unsigned_type (type_node)
      tree type_node;
 {
-  return type_for_size (TYPE_PRECISION (type_node), 1);
+  tree type = type_for_size (TYPE_PRECISION (type_node), 1);
+
+  if (TYPE_MODULAR_P (type_node))
+    {
+      type = copy_node (type);
+      TREE_TYPE (type) = type_node;
+    }
+  else if (TREE_TYPE (type_node) != 0
+	   && TYPE_MODULAR_P (TREE_TYPE (type_node)))
+    {
+      type = copy_node (type);
+      TREE_TYPE (type) = TREE_TYPE (type_node);
+    }
+
+  return type;
 }
 
 /* Return the signed version of a TYPE_NODE, a scalar type.  */
@@ -1232,7 +1388,21 @@ tree
 signed_type (type_node)
      tree type_node;
 {
-  return type_for_size (TYPE_PRECISION (type_node), 0);
+  tree type = type_for_size (TYPE_PRECISION (type_node), 0);
+
+  if (TYPE_MODULAR_P (type_node))
+    {
+      type = copy_node (type);
+      TREE_TYPE (type) = type_node;
+    }
+  else if (TREE_TYPE (type_node) != 0
+	   && TYPE_MODULAR_P (TREE_TYPE (type_node)))
+    {
+      type = copy_node (type);
+      TREE_TYPE (type) = TREE_TYPE (type_node);
+    }
+
+  return type;
 }
 
 /* Return a type the same as TYPE except unsigned or signed according to
@@ -1357,16 +1527,65 @@ build_template (template_type, array_type, expr)
   return build_constructor (template_type, nreverse (template_elts));
 }
 
+/* Update anything previously pointing to OLD_TYPE to point to NEW_TYPE.  In
+   the normal case this is just two adjustments, but we have more to do
+   if NEW is an UNCONSTRAINED_ARRAY_TYPE.  */
+
+void
+update_pointer_to (old_type, new_type)
+     tree old_type;
+     tree new_type;
+{
+  tree ptr = TYPE_POINTER_TO (old_type);
+
+  if (ptr == 0)
+    return;
+
+  /* First handle the simple case.  */
+  if (TREE_CODE (new_type) != UNCONSTRAINED_ARRAY_TYPE)
+    {
+      TREE_TYPE (ptr) = new_type;
+      TYPE_POINTER_TO (new_type) = ptr;
+    }
+
+  /* Now deal with the unconstrained array case. In this case the "pointer"
+     is actually a RECORD_TYPE where the types of both fields are
+     pointers to void.  In that case, copy the field list from the
+     old type to the new one and update the fields' context. */
+  else if (TREE_CODE (ptr) != RECORD_TYPE || ! TYPE_FAT_POINTER_P (ptr))
+    gigi_abort (412);
+
+  else
+    {
+      tree ptr_temp_type;
+
+      TYPE_FIELDS (ptr) = TYPE_FIELDS (TYPE_POINTER_TO (new_type));
+      DECL_CONTEXT (TYPE_FIELDS (ptr)) = ptr;
+      DECL_CONTEXT (TREE_CHAIN (TYPE_FIELDS (ptr))) = ptr;
+
+      /* Rework the PLACEHOLDER_EXPR inside the reference to the
+	 template bounds.  */
+      ptr_temp_type = TREE_TYPE (TREE_CHAIN (TYPE_FIELDS (ptr)));
+      update_pointer_to 
+	(TREE_TYPE (TREE_TYPE (TYPE_FIELDS (ptr))),
+	 gnat_substitute_in_type (TREE_TYPE (TREE_TYPE (TYPE_FIELDS (ptr))),
+				  TREE_CHAIN (TYPE_FIELDS (ptr)),
+				  build (COMPONENT_REF, ptr_temp_type,
+					 build (PLACEHOLDER_EXPR, ptr),
+					 TREE_CHAIN (TYPE_FIELDS (ptr)))));
+
+      TYPE_UNCONSTRAINED_ARRAY (ptr) = new_type;
+      TYPE_POINTER_TO (new_type) = TREE_TYPE (new_type) = ptr;
+    }
+}
+
 /* Convert a pointer to a constrained array into a pointer to an unconstrained
-   array.  This involves making a template.  If ALLOCATE is nonzero, the
-   template must be created using an allocator instead of in the
-   current context's memory.  */
+   array.  This involves making a template.  */
 
 tree
-convert_to_unconstrained (type, expr, allocate)
+convert_to_unconstrained (type, expr)
      tree type;
      tree expr;
-     int allocate;
 {
   tree template_type = TREE_TYPE (TREE_TYPE (TREE_CHAIN (TYPE_FIELDS (type))));
   tree template_cons, template_addr;
@@ -1384,17 +1603,25 @@ convert_to_unconstrained (type, expr, allocate)
 					expr),
 			       NULL_TREE)));
 
-  /* Build the constructor for the template.  */
-  template_cons
-    = build_template (template_type, TREE_TYPE (TREE_TYPE (expr)), expr);
+  /* Build the constructor for the template.  If EXPR is a pointer to a
+     COMPONENT_REF from a TYPE_CONTAINS_TEMPLATE_P, get the template from the
+     underlying type.  In that case, we don't have to worry about
+     allocation.  */
+  if (TREE_CODE (expr) == ADDR_EXPR
+      && TREE_CODE (TREE_OPERAND (expr, 0)) == COMPONENT_REF
+      && (TYPE_CONTAINS_TEMPLATE_P
+	  (TREE_TYPE (TREE_OPERAND (TREE_OPERAND (expr, 0), 0)))))
+    {
+      tree underlying = TREE_OPERAND (TREE_OPERAND (expr, 0), 0);
+      tree temp_field = TYPE_FIELDS (TREE_TYPE (underlying));
 
-  /* If the template needs to be allocated in the heap, allocate memory for its
-     constructor.  Otherwise, just take the address of the constructor.  */
-  if (allocate)
-    template_addr = build_allocator (template_type, template_cons,
-				     build_pointer_type (template_type));
+      template_cons = build_component_ref (underlying, NULL_TREE, temp_field);
+    }
   else
-    template_addr = build_unary_op (ADDR_EXPR, NULL_TREE, template_cons);
+    template_cons
+      = build_template (template_type, TREE_TYPE (TREE_TYPE (expr)), expr);
+
+  template_addr = build_unary_op (ADDR_EXPR, NULL_TREE, template_cons);
 
   /* The result is a CONSTRUCTOR for the fat pointer.  */
   return
@@ -1415,9 +1642,11 @@ convert (type, expr)
      tree type, expr;
 {
   enum tree_code code = TREE_CODE (type);
+  tree etype = TREE_TYPE (expr);
+  enum tree_code ecode = TREE_CODE (etype);
 
   /* If EXPR is already the right type, we are done.  */
-  if (type == TREE_TYPE (expr))
+  if (type == etype)
     return expr;
 
   /* There are some special cases of expressions that we process
@@ -1436,7 +1665,7 @@ convert (type, expr)
     case STRING_CST:
       /* If we are converting a STRING_CST to another constrained array type,
 	 just make a new one in the proper type.  */
-      if (TREE_CODE (type) != ARRAY_TYPE)
+      if (code != ARRAY_TYPE)
 	break;
 
       expr = copy_node (expr);
@@ -1457,11 +1686,13 @@ convert (type, expr)
      a constrained array.  */
   if (TYPE_FAT_POINTER_P (type)
       && (integer_zerop (expr)
-	  || (TREE_CODE (TREE_TYPE (expr)) == POINTER_TYPE
-	      && TREE_CODE (TREE_TYPE (TREE_TYPE (expr))) == ARRAY_TYPE)))
-    return convert_to_unconstrained (type, expr, 0);
+	  || (ecode == POINTER_TYPE
+	      && TREE_CODE (TREE_TYPE (etype)) == ARRAY_TYPE)))
+    return convert_to_unconstrained (type, expr);
 
-  if (TYPE_MAIN_VARIANT (type) == TYPE_MAIN_VARIANT (TREE_TYPE (expr)))
+  if (TYPE_MAIN_VARIANT (type) == TYPE_MAIN_VARIANT (etype)
+      || (code == INTEGER_CST && ecode == INTEGER_CST
+	  && (type == TREE_TYPE (etype) || etype == TREE_TYPE (type))))
     return fold (build1 (NOP_EXPR, type, expr));
 
   switch (code)
@@ -1470,9 +1701,18 @@ convert (type, expr)
       return build1 (CONVERT_EXPR, type, expr);
 
     case INTEGER_TYPE:
-      /* If TYPE has a modulus, first convert to its base type,
-	 then take the modulus and convert to TYPE.  */
-      if (TREE_TYPE (type) != 0 && TYPE_MODULAR_P (TREE_TYPE (type)))
+      /* If TYPE has a modulus and EXPR's type does not (or has a 
+	 different modulus), first convert to its base type, then take the
+	 modulus and convert to TYPE.  */
+      if (TREE_TYPE (type) != 0 && TYPE_MODULAR_P (TREE_TYPE (type))
+	  && ! (ecode == INTEGER_TYPE && TYPE_MODULAR_P (etype)
+		&& tree_int_cst_equal (TYPE_MODULUS (TREE_TYPE (type)),
+				       TYPE_MODULUS (etype)))
+	  && ! (ecode == INTEGER_TYPE
+		&& TREE_TYPE (etype) != 0
+		&& TYPE_MODULAR_P (TREE_TYPE (etype))
+		&& tree_int_cst_equal (TYPE_MODULUS (TREE_TYPE (etype)),
+				       TYPE_MODULUS (TREE_TYPE (type)))))
 	{
 	  tree base_type = TREE_TYPE (type);
 	  tree result = fold (convert_to_integer (base_type, expr));
@@ -1489,6 +1729,12 @@ convert (type, expr)
       return fold (convert_to_integer (type, expr));
 
     case POINTER_TYPE:
+      /* If converting fat pointer to normal pointer, get the pointer to the
+	 array and then convert it.  */
+      if (TYPE_FAT_POINTER_P (etype))
+	expr = build_component_ref (expr, get_identifier ("P_ARRAY"),
+				    NULL_TREE);
+
       return fold (convert_to_pointer (type, expr));
 
     case REAL_TYPE:
@@ -1499,48 +1745,41 @@ convert (type, expr)
       /* In these cases, assume the front-end has validated the conversion.
 	 If the conversion is valid, it will bit a bit-wise conversion, so
 	 it can be viewed as an unchecked conversion.  */
-      return build1 (UNCHECKED_CONVERT_EXPR, type, expr);
+      expr = build1 (UNCHECKED_CONVERT_EXPR, type, expr);
+
+      /* If our result has side-effects and is of an unconstrained type,
+	 make a SAVE_EXPR so that we can be sure it will only be referenced
+	 once.  */
+      if (TREE_SIDE_EFFECTS (expr)
+	  && (TREE_CODE (type) == UNCONSTRAINED_ARRAY_TYPE
+	      || (TREE_CODE (TYPE_SIZE (type)) != INTEGER_CST
+		  && contains_placeholder_p (TYPE_SIZE (type)))))
+	expr = make_save_expr (expr);
+      return expr;
 
     case UNCONSTRAINED_ARRAY_TYPE:
-      /* In this case EXPR must be a constrained array.  In that case,
-	 take its address, convert it to a fat pointer, and then
-	 dereference it.  */
-      if (TREE_CODE (TREE_TYPE (expr)) != ARRAY_TYPE)
+      /* If EXPR is a constrained array, take its address, convert it to a
+	 fat pointer, and then dereference it.  */
+      if (ecode == ARRAY_TYPE)
+	return
+	  build_unary_op
+	    (INDIRECT_REF, NULL_TREE,
+	     convert_to_unconstrained (TREE_TYPE (type),
+				       build_unary_op (ADDR_EXPR,
+						       NULL_TREE, expr)));
+
+      /* Do something very similar for converting one unconstrained
+	 array to another.  */
+      else if (ecode == UNCONSTRAINED_ARRAY_TYPE)
+	return
+	  build_unary_op (INDIRECT_REF, NULL_TREE,
+			  convert (TREE_TYPE (type),
+				   build_unary_op (ADDR_EXPR,
+						   NULL_TREE, expr)));
+      else
 	gigi_abort (409);
 
-      return
-	build_unary_op
-	  (INDIRECT_REF, NULL_TREE,
-	   convert_to_unconstrained (TREE_TYPE (type),
-				     build_unary_op (ADDR_EXPR,
-						     NULL_TREE, expr),
-				     0));
     default:
       gigi_abort (410);
     }
 }
-
-#if 0
-----------------------
--- REVISION HISTORY --
-----------------------
-
-----------------------------
-revision 1.86
-date: Tue May 17 08:32:44 1994;  author: kenner
-(build_template): New function.
-(convert_to_unconstrained): Use it.
-----------------------------
-revision 1.87
-date: Mon Jul 11 07:25:11 1994;  author: kenner
-(create_subprog_type): New arg, RETURNS_UNCONSTRAINED.
-Make new type copy if RETURNED_UNCONSTRAINED differs from value in type.
-Set TYPE_RETURNS_UNCONSTRAINED_P here.
-----------------------------
-revision 1.88
-date: Mon Jul 18 16:36:59 1994;  author: banner
-(convert): change p_array to P_ARRAY and p_template to P_TEMPLATE.
- change requested by Paul for interaction with gdb.
-----------------------------
-** New changes after this line and before endif. **
-#endif
