@@ -6,7 +6,7 @@
 /*                                                                          */
 /*                              C Header File                               */
 /*                                                                          */
-/*                            $Revision: 1.53 $                             */
+/*                            $Revision: 1.73 $                             */
 /*                                                                          */
 /*           Copyright (c) 1992,1993,1994 NYU, All Rights Reserved          */
 /*                                                                          */
@@ -22,6 +22,9 @@
 /*                                                                          */
 /****************************************************************************/
 
+/* If nonzero, pretend we are allocating at global level.  */
+extern int force_global;
+
 /* Standard data type sizes.  Most of these are not used.  */
 
 #ifndef CHAR_TYPE_SIZE
@@ -66,21 +69,41 @@
 /* Nonzero if this type is a record corresponding to a fat pointer.  */
 #define TYPE_FAT_POINTER_P(NODE) TYPE_LANG_FLAG_0 (NODE)
 
-/* Nonzero if this is a modular type with a modulus that is not equal to
-   two to the power of its mode's size.  */
+/* For INTEGER_TYPE, nonzero if this is a modular type with a modulus that
+   is not equal to two to the power of its mode's size.  */
 #define TYPE_MODULAR_P(NODE) TYPE_LANG_FLAG_1 (NODE)
 
-/* Nonzero if this is a FUNCTION_TYPE that denotes a function returning
-   an unconstrained array or record.  */
-#define TYPE_RETURNS_UNCONSTRAINED_P(NODE) TYPE_LANG_FLAG_2 (NODE)
-
-/* Nonzero in an arithmetic subtype if the fact that this is a subtype
-   is not known to the front-end.  */
-#define TYPE_EXTRA_SUBTYPE_P(NODE) TYPE_LANG_FLAG_3 (NODE)
-
-/* Nonzero in an ARRAY_TYPE if this type corresponds to a dimension of
+/* For ARRAY_TYPE, nonzero if this type corresponds to a dimension of
    an Ada array other than the first.  */
-#define TYPE_MULTI_ARRAY_P(NODE)  TYPE_LANG_FLAG_4 (NODE)
+#define TYPE_MULTI_ARRAY_P(NODE)  TYPE_LANG_FLAG_1 (NODE)
+
+/* For FUNCTION_TYPE, nonzero if this denotes a function returning an
+   unconstrained array or record.  */
+#define TYPE_RETURNS_UNCONSTRAINED_P(NODE) TYPE_LANG_FLAG_1 (NODE)
+
+/* For RECORD_TYPE, nonzero if this type has a record rep clause.  */
+#define TYPE_HAS_REP_CLAUSE_P(NODE)  TYPE_LANG_FLAG_1 (NODE)
+
+/* Nonzero in an arithmetic subtype if this is a subtype not known to the
+   front-end.  */
+#define TYPE_EXTRA_SUBTYPE_P(NODE) TYPE_LANG_FLAG_2 (NODE)
+
+/* Nonzero if this is the type for an object whose type includes
+   its template in addition to its value.  */
+#define TYPE_CONTAINS_TEMPLATE_P(NODE) TYPE_LANG_FLAG_3 (NODE)
+
+/* For ARRAY_TYPEs, nonzero if the array type has Convention_Fortran.  */
+#define TYPE_CONVENTION_FORTRAN_P(NODE) TYPE_LANG_FLAG_4 (NODE)
+
+/* For VOID_TYPE and RECORD_TYPE, nonzero if this is a dummy type,
+   make to correspond to a private or incomplete type.  */
+#define TYPE_DUMMY_P(NODE) TYPE_LANG_FLAG_4 (NODE)
+
+/* True if TYPE is such a dummy type.  */
+#define TYPE_IS_DUMMY_P(NODE) \
+  ((TREE_CODE (NODE) == VOID_TYPE || TREE_CODE (NODE) == RECORD_TYPE) \
+   && TYPE_DUMMY_P (NODE))
+
 /* This field is only defined for FUNCTION_TYPE nodes. If the Ada
    subprogram contains no parameters passed by copy in/copy out then this
    field is 0. Otherwise it points to a list of nodes used to specify the
@@ -115,6 +138,20 @@
 /* Nonzero in a FIELD_DECL that is a parent field.  */
 #define DECL_PARENT_P(NODE) DECL_LANG_FLAG_2 (NODE)
 
+/* Nonzero in a PARM_DECL if this was an Ada array being passed to a
+   foreign convention subprogram.  */
+#define DECL_BY_COMPONENT_PTR_P(NODE) DECL_LANG_FLAG_3 (NODE)
+
+/* Nonzero in a FIELD_DECL that is the first field in a variant of
+   record with a record rep clause.  */
+#define DECL_FIRST_FIELD_IN_VARIANT_P(NODE) DECL_LANG_FLAG_4 (NODE)
+
+/* Nonzero in a FIELD_DECL that is a dummy built for a variant part.  */
+#define DECL_FOR_VARIANT_P(NODE) DECL_LANG_FLAG_5 (NODE)
+
+/* Nonzero in a TRANSFORM_EXPR if we are to take its address.  */
+#define TREE_TRANSFORM_ADDR(NODE) TREE_STATIC (NODE)
+
 /* Variables expected by the GCC back-end.  */
 
 /* A node which has tree code ERROR_MARK, and whose type is itself.  */
@@ -132,6 +169,9 @@ extern tree integer_zero_node;
 extern tree integer_one_node;
 extern tree null_pointer_node;
 
+/* The type of an exception.  */
+extern tree except_type_node;
+
 /* The FUNCTION_DECL node for the function currently being compiled, or 0
    if between functions.  */
 extern tree current_function_decl;
@@ -152,6 +192,9 @@ extern tree ptr_void_ftype;
    Ada allocators cause calls to this function to be generated.   */
 extern tree malloc_decl;
 
+/* Likewise for freeing memory.  */
+extern tree free_decl;
+
 /* Types and decls used by our temporary exception mechanism.  See
    init_decl_processing for details.  */
 extern tree jmpbuf_type;
@@ -161,7 +204,6 @@ extern tree set_jmpbuf_decl;
 extern tree get_excptr_decl;
 extern tree raise_decl;
 extern tree raise_nodefer_decl;
-extern tree reraise_decl;
 extern tree setjmp_decl;
 extern tree longjmp_decl;
 extern tree raise_constraint_error_decl;
@@ -276,8 +318,10 @@ extern int present_gnu_tree	PROTO((Entity_Id));
 extern void init_gnat_to_gnu		PROTO((void));
 
 /* Given a record type (RECORD_TYPE) and a chain of FIELD_DECL
-   nodes (FIELDLIST), finish constructing the record or union type.  */
-extern void finish_record_type		PROTO((tree, tree));
+   nodes (FIELDLIST), finish constructing the record or union type. 
+   If HAS_REP is nonzero, this record has a rep clause; don't call
+   layout_type but merely set the size and alignment ourselves.  */
+extern void finish_record_type		PROTO((tree, tree, int));
 
 /* Returns a FUNCTION_TYPE node. RETURN_TYPE is the type returned by the
    subprogram. If it is void_type_node, then we are dealing with a procedure,
@@ -306,6 +350,9 @@ extern tree create_type_decl		PROTO((char *, tree));
    its data type (a GCC ..._TYPE node).  VAR_INIT is the GCC tree for an
    optional initial expression; NULL_TREE if none.
 
+   SIZE, if nonzero, is a GCC tree to be used for the size of the variable.
+   ALIGN, if nonzero, is the required alignment of the variable.
+
    CONST_FLAG is nonzero if this variable is constant.
 
    PUBLIC_FLAG is nonzero if this definition is to be made visible outside of
@@ -315,7 +362,7 @@ extern tree create_type_decl		PROTO((char *, tree));
    definition: no storage is to be allocated for the variable here).
    STATIC_FLAG is only relevant when not at top level.  In that case
    it indicates whether to always allocate storage to the variable.  */
-extern tree create_var_decl	PROTO((char *, char *, tree, tree,
+extern tree create_var_decl	PROTO((char *, char *, tree, tree, tree, int,
 				       int, int, int, int));
 
 /* Obtain any pending elaborations and clear the old list.  */
@@ -327,8 +374,9 @@ extern void add_pending_elaborations PROTO ((tree, tree));
 
 /* Returns a FIELD_DECL node. FIELD_NAME the field name, FIELD_TYPE is its
    type, and RECORD_TYPE is the type of the parent.  PACKED is nonzero if
-   this field is in a record type with a "pragma pack".  */
-extern tree create_field_decl	PROTO((char *, tree, tree, int));
+   this field is in a record type with a "pragma pack".  If SIZE is nonnegative
+   it is the specified size for this field and POS is the bit position.  */
+extern tree create_field_decl	PROTO((char *, tree, tree, int, int, int));
 
 /* Returns a PARM_DECL node. PARAM_NAME is the name of the parameter,
    PARAM_TYPE is its type.  */
@@ -342,7 +390,7 @@ extern tree create_param_decl	PROTO((char *, tree));
    INLINE_FLAG, PUBLIC_FLAG, EXTERN_FLAG, and PURE_FLAG are used to set the
    appropriate fields in the FUNCTION_DECL.  */
 extern tree create_subprog_decl	PROTO((char *, char *, tree, tree,
-				       int, int, int, int));
+				       int, int, int, int, tree));
 
 /* Returns a LABEL_DECL node for LABEL_NAME.  */
 extern tree create_label_decl	PROTO((char *));
@@ -361,34 +409,16 @@ extern void end_subprog_body	PROTO((void));
    Return a constructor for the template.  */
 extern tree build_template	PROTO((tree, tree, tree));
 
+/* Update anything previously pointing to OLD_TYPE to point to NEW_TYPE.  In
+   the normal case this is just two adjustments, but we have more to do
+   if NEW is an UNCONSTRAINED_ARRAY_TYPE.  */
+extern void update_pointer_to	PROTO((tree, tree));
+
 /* Convert a pointer to a constrained array into a pointer to an unconstrained
-   array.  This involves making a template.  If ALLOCATE is nonzero, the
-   template must be created using an allocator instead of in the
-   current context's memory.  */
-extern tree convert_to_unconstrained PROTO((tree, tree, int));
+   array.  This involves making a template.  */
+extern tree convert_to_unconstrained PROTO((tree, tree));
 
 /* EXP is an expression for the size of an object.  If this size contains
    discriminant references, replace them with the maximum (if MAX_P) or
    minimum (if ! MAX_P) possible value of the discriminant.  */
 extern tree max_size		PROTO((tree, int));
-
-#if 0
-----------------------
--- REVISION HISTORY --
-----------------------
-
-----------------------------
-revision 1.51
-date: Tue May 17 08:32:52 1994;  author: kenner
-(build_template): New definition.
-----------------------------
-revision 1.52
-date: Sun Jun 12 16:30:49 1994;  author: kenner
-(TYPE_MULTI_ARRAY_P): New flag.
-----------------------------
-revision 1.53
-date: Mon Jul 11 07:25:19 1994;  author: kenner
-(create_subprog_type): Add new parms, RETURNS_UNCONSTRAINED.
-----------------------------
-** New changes after this line and before endif. **
-#endif
