@@ -6,7 +6,7 @@
 /*                                                                          */
 /*                          C Implementation File                           */
 /*                                                                          */
-/*                            $Revision: 1.74 $                             */
+/*                            $Revision: 1.91 $                             */
 /*                                                                          */
 /*           Copyright (c) 1992,1993,1994 NYU, All Rights Reserved          */
 /*                                                                          */
@@ -36,6 +36,7 @@
 #include "a-snames.h"
 #include "a-string.h"
 #include "a-uintp.h"
+#include "a-gtran3.h"
 #include "a-trans.h"
 #include "a-trans3.h"
 #include "a-trans4.h"
@@ -86,7 +87,7 @@ truthvalue_conversion (expr)
 
 /* Return the base type of TYPE.  */
 
-static tree
+tree
 get_base_type (type)
      tree type;
 {
@@ -97,7 +98,6 @@ get_base_type (type)
 
   return type;
 }
-
 
 /* We have a comparison or assignment operation on two array types, T1 and T2.
    Return the type that both operands should be converted to, if any.
@@ -252,6 +252,105 @@ compare_arrays (result_type, a1, a2)
   return result;
 }
 
+/* Compute the result of applying OP_CODE to LHS and RHS, where both are of
+   type TYPE.  We know that TYPE is a modular type with a nonbinary
+   modulus.  */
+
+static tree
+nonbinary_modular_operation (op_code, type, lhs, rhs)
+     enum tree_code op_code;
+     tree type;
+     tree lhs, rhs;
+{
+  tree modulus = TYPE_MODULUS (type);
+  int needed_precision
+    = (TREE_INT_CST_HIGH (modulus) != 0
+       ? HOST_BITS_PER_WIDE_INT + floor_log2 (TREE_INT_CST_HIGH (modulus)) + 1
+       : TREE_INT_CST_LOW (modulus) == 0 ? 1
+       : floor_log2 (TREE_INT_CST_LOW (modulus)) + 1);
+  int precision;
+  int unsignedp = 1;
+  tree op_type = type;
+  tree result;
+
+  /* If this is an addition of a constant, convert it to a subtraction
+     of a constant since we can do that faster.  */
+  if (op_code == PLUS_EXPR && TREE_CODE (rhs) == INTEGER_CST)
+    rhs = fold (build (MINUS_EXPR, type, modulus, rhs)), op_code = MINUS_EXPR;
+
+  /* For the logical operations, we only need PRECISION bits.  For
+     addition and subraction, we need one more and for multiplication we
+     need twice as many.  But we never want to make a size smaller than
+     our size. */
+  if (op_code == PLUS_EXPR || op_code == MINUS_EXPR)
+    needed_precision += 1;
+  else if (op_code == MULT_EXPR)
+    needed_precision *= 2;
+
+  precision = MAX (needed_precision, TYPE_PRECISION (op_type));
+
+  /* Unsigned will do for everything but subtraction.  */
+  if (op_code == MINUS_EXPR)
+    unsignedp = 0;
+
+  /* If our type is the wrong signedness or isn't wide enough, make a new
+     type and convert both our operands to it.  */
+  if (TYPE_PRECISION (op_type) < precision
+      || TREE_UNSIGNED (op_type) != unsignedp)
+    {
+      /* Copy the node so we ensure it can be modified to make it modular.  */
+      op_type = copy_node (type_for_size (precision, unsignedp));
+      modulus = convert (op_type, modulus);
+      TYPE_MODULUS (op_type) = modulus;
+      TYPE_MODULAR_P (op_type) = 1;
+      lhs = convert (op_type, lhs);
+      rhs = convert (op_type, rhs);
+    }
+
+  /* Do the operation, then we'll fix it up.  */
+  result = fold (build (op_code, op_type, lhs, rhs));
+
+  /* For multiplication, we have no choice but to do a full modulus
+     operation.  However, we want to do this in the narrowest
+     possible size.  */
+  if (op_code == MULT_EXPR)
+    {
+      tree div_type = copy_node (type_for_size (needed_precision, 1));
+      modulus = convert (div_type, modulus);
+      TYPE_MODULUS (div_type) = modulus;
+      TYPE_MODULAR_P (div_type) = 1;
+      result = convert (op_type,
+			fold (build (TRUNC_MOD_EXPR, div_type,
+				     convert (div_type, result), modulus)));
+    }
+
+  /* For subtraction, add the modulus back if we are negative.  */
+  else if (op_code == MINUS_EXPR)
+    {
+      result = save_expr (result);
+      result = fold (build (COND_EXPR, op_type,
+			    build (LT_EXPR, integer_type_node, result,
+				   convert (op_type, integer_zero_node)),
+			    fold (build (PLUS_EXPR, op_type,
+					 result, modulus)),
+			    result));
+    }
+
+  /* For the other operations, subtract the modulus if we are >= it.  */
+  else
+    {
+      result = save_expr (result);
+      result = fold (build (COND_EXPR, op_type,
+			    build (GE_EXPR, integer_type_node,
+				   result, modulus),
+			    fold (build (MINUS_EXPR, op_type,
+					 result, modulus)),
+			    result));
+    }
+
+  return convert (type, result);
+}
+
 /* Make a binary operation of kind OP_CODE.  RESULT_TYPE is the type
    desired for the result.  Usually the operation is to be performed
    in that type.  For MODIFY_EXPR and ARRAY_REF, RESULT_TYPE may be 0
@@ -313,6 +412,10 @@ build_binary_op (op_code, result_type, left_operand, right_operand)
       if (right_type != TYPE_DOMAIN (left_type))
 	right_operand = convert (TYPE_DOMAIN (left_type), right_operand);
 
+      if (! TREE_CONSTANT (right_operand)
+	  || ! TREE_CONSTANT (TYPE_MIN_VALUE (right_type)))
+	mark_addressable (left_operand);
+
       modulus = 0;
       break;
 
@@ -327,7 +430,7 @@ build_binary_op (op_code, result_type, left_operand, right_operand)
 
     case EQ_EXPR:
     case NE_EXPR:
-      /* If both objects are records, compare them specially.  */
+      /* If both objects are arrays, compare them specially.  */
       if (TREE_CODE (left_type) == ARRAY_TYPE
 	  && TREE_CODE (right_type) == ARRAY_TYPE)
 	{
@@ -443,8 +546,28 @@ build_binary_op (op_code, result_type, left_operand, right_operand)
     case TRUTH_XOR_EXPR:
       left_operand = truthvalue_conversion (left_operand);
       right_operand = truthvalue_conversion (right_operand);
-      /* ... fall through ... */
+      goto common;
+
+    case BIT_AND_EXPR:
+    case BIT_IOR_EXPR:
+    case BIT_XOR_EXPR:
+      /* For binary modulus, if the inputs are in range, so are the
+	 outputs.  */
+      if (modulus != 0 && integer_pow2p (modulus))
+	modulus = 0;
+
+      goto common;
+
+    case TRUNC_DIV_EXPR:   case TRUNC_MOD_EXPR:
+    case CEIL_DIV_EXPR:    case CEIL_MOD_EXPR:
+    case FLOOR_DIV_EXPR:   case FLOOR_MOD_EXPR:
+    case ROUND_DIV_EXPR:   case ROUND_MOD_EXPR:
+      /* These always produce results lower than either operand.  */
+      modulus = 0;
+      goto common;
+
     default:
+    common:
       /* The result type should be the same as the base types of the
 	 both operands (and they should be the same).  Convert
 	 everything to the result type.  */
@@ -457,23 +580,23 @@ build_binary_op (op_code, result_type, left_operand, right_operand)
       right_operand = convert (operation_type, right_operand);
     }
 
-  /* For non-binary modular types, do the operation as signed so that we get
-     mathematically correct answer.  */
-  if (modulus != 0 && TREE_UNSIGNED (operation_type))
+  if (modulus != 0 && ! integer_pow2p (modulus))
     {
-      operation_type = signed_type (operation_type);
-      left_operand = convert (operation_type, left_operand);
-      right_operand = convert (operation_type, right_operand);
+      result = nonbinary_modular_operation (op_code, operation_type,
+					    left_operand, right_operand);
+      modulus = 0;
     }
+  else
+    result = fold (build (op_code, operation_type,
+			  left_operand, right_operand));
 
-  result = build (op_code, operation_type, left_operand, right_operand);
-  TREE_SIDE_EFFECTS (result) = has_side_effects;
-
-  result = fold (result);
+  TREE_SIDE_EFFECTS (result) |= has_side_effects;
   TREE_CONSTANT (result)
-    = TREE_CONSTANT (left_operand) & TREE_CONSTANT (right_operand);
+    = (TREE_CONSTANT (left_operand) & TREE_CONSTANT (right_operand)
+       && op_code != ARRAY_REF);
 
-  /* If we are working with modular types, perform the MOD operation.  */
+  /* If we are working with modular types, perform the MOD operation
+     if something above hasn't eliminated the need for it.  */
   if (modulus != 0)
     result = fold (build (FLOOR_MOD_EXPR, operation_type, result,
 			  convert (operation_type, modulus)));
@@ -496,8 +619,6 @@ build_unary_op (op_code, result_type, operand)
   tree base_type = get_base_type (type);
   tree operation_type = (result_type != 0 && TYPE_EXTRA_SUBTYPE_P (result_type)
 			 ? get_base_type (result_type) : result_type);
-  tree modulus = (operation_type && TYPE_MODULAR_P (operation_type)
-		  ? TYPE_MODULUS (operation_type) : 0);
   tree result;
   int side_effects = 0;
 
@@ -514,16 +635,31 @@ build_unary_op (op_code, result_type, operand)
       if (TREE_CODE (operand) == INDIRECT_REF
 	  || TREE_CODE (operand) == UNCONSTRAINED_ARRAY_REF)
 	result = TREE_OPERAND (operand, 0);
+      else if (TREE_CODE (operand) == TRANSFORM_EXPR)
+	{
+	  TREE_TRANSFORM_ADDR (operand) = 1;
+	  result = operand;
+
+	  if (type != error_mark_node)
+	    {
+	      if (TREE_CODE (type) == UNCONSTRAINED_ARRAY_TYPE)
+		type = TREE_TYPE (type);
+	      else
+		type = build_pointer_type (type);
+
+	      TREE_TYPE (result) = type;
+	    }
+	}
       else
 	{
 	  if (type != error_mark_node)
 	    operation_type = build_pointer_type (type);
 
+	  mark_addressable (operand);
 	  result = fold (build1 (op_code, operation_type, operand));
 	}
 
       TREE_CONSTANT (result) = staticp (operand) || TREE_CONSTANT (operand);
-      modulus = 0;
       break;
 
     case INDIRECT_REF:
@@ -536,23 +672,89 @@ build_unary_op (op_code, result_type, operand)
       else
 	result = fold (build1 (op_code, TREE_TYPE (type), operand));
 
-      modulus = 0;
       side_effects = flag_volatile 
 	|| (! TYPE_FAT_POINTER_P (type) && TYPE_VOLATILE (TREE_TYPE (type)));
       break;
 
+    case NEGATE_EXPR:
+    case BIT_NOT_EXPR:
+      {
+	tree modulus = (operation_type && TYPE_MODULAR_P (operation_type)
+			? TYPE_MODULUS (operation_type) : 0);
+	int mod_pow2 = modulus != 0 && integer_pow2p (modulus);
+
+	/* If this is a modular type, there are various possibilities
+	   depending on the operation and whether the modulus is a
+	   power of two or not.  */
+
+	if (modulus != 0)
+	  {
+	    if (operation_type != base_type)
+	      gigi_abort (509);
+
+	    operand = convert (operation_type, operand);
+
+	    /* The fastest in the negate case for binary modulus is
+	       the straightforward code; the TRUNC_MOD_EXPR below
+	       is an AND operation.  */
+	    if (op_code == NEGATE_EXPR && mod_pow2)
+	      result = fold (build (TRUNC_MOD_EXPR, operation_type,
+				    fold (build1 (NEGATE_EXPR, operation_type,
+						  operand)),
+				    modulus));
+
+	    /* For nonbinary negate case, return zero for zero operand,
+	       else return the modulus minus the operand.  If the modulus
+	       is a power of two minus one, we can do the subtraction
+	       as an XOR since it is equivalent and faster on most machines. */
+	    else if (op_code == NEGATE_EXPR && ! mod_pow2)
+	      {
+		if (integer_pow2p (fold (build (PLUS_EXPR, operation_type,
+						modulus,
+						convert (operation_type,
+							 integer_one_node)))))
+		  result = fold (build (BIT_XOR_EXPR, operation_type,
+					operand, modulus));
+		else
+		  result = fold (build (MINUS_EXPR, operation_type,
+					modulus, operand));
+
+		result = fold (build (COND_EXPR, operation_type,
+				      fold (build (NE_EXPR, integer_type_node,
+						   operand,
+						   convert (operation_type,
+							    integer_zero_node))),
+				      result, operand));
+	      }
+	    else
+	      {
+		/* For the NOT cases, we need a constant equal to
+		   the modulus minus one.  For a binary modulus, we
+		   XOR against the constant and subtract the operand from
+		   that constant for nonbinary modulus.  */
+
+		tree cnst = fold (build (MINUS_EXPR, operation_type, modulus,
+					 convert (operation_type,
+						  integer_one_node)));
+
+		if (mod_pow2)
+		  result = fold (build (BIT_XOR_EXPR, operation_type,
+					operand, cnst));
+		else
+		  result = fold (build (MINUS_EXPR, operation_type,
+					cnst, operand));
+	      }
+
+	    break;
+	  }
+      }
+
+      /* ... fall through ... */
+
     default:
       if (operation_type != base_type)
 	gigi_abort (509);
 
-      /* For non-binary modular types, do the operation as signed so that we
-	 get mathematically correct answer.  */
-      if (modulus != 0 && TREE_UNSIGNED (operation_type))
-	{
-	  operation_type = signed_type (operation_type);
-	  operand = convert (operation_type, operand);
-	}
-
       result = fold (build1 (op_code, operation_type, convert (operation_type,
 							       operand)));
     }
@@ -567,11 +769,6 @@ build_unary_op (op_code, result_type, operand)
   if (result_type != 0 && TREE_TYPE (result) != result_type)
     result = convert (result_type, result);
 
-  /* If we are working with modular types, perform the MOD operation.  */
-  if (modulus != 0)
-    result = fold (build (FLOOR_MOD_EXPR, operation_type, result,
-			  convert (operation_type, modulus)));
-
   return result;
 }
 
@@ -701,7 +898,7 @@ build_simple_component_ref (record_variable, component, field)
 	{
 	  if (DECL_NAME (field) == component)
 	    break;
-	  else if (DECL_NAME (field) == 0)
+	  else if (DECL_FOR_VARIANT_P (field))
 	     {
 	      tree field_ref
 		= build_simple_component_ref (record_variable, 
@@ -756,20 +953,95 @@ build_component_ref (record_variable, component, field)
 					   parent_comp, NULL_TREE);
   if (parent_ref)
     return build_component_ref (parent_ref, component, field);
+
+  /* If FIELD was specified, assume this is an invalid user field so
+     raise constraint error.  Otherwise, we can't find the type to return, so
+     abort.  */
+  else if (field != 0)
+    return build1 (NULL_EXPR, TREE_TYPE (field),
+		   build_call_0_expr (raise_constraint_error_decl));
   else
-    /* The field was not found in the hierarchy. Should not happen. */
     gigi_abort (512);
 }
 
+/* Build a GCC tree to call an allocation or deallocation function.
+   If GNU_OBJ is nonzero, it is an object to deallocate.  Otherwise,
+   genrate an allocator.
+
+   GNU_SIZE is the size of the object and ALIGN is the alignment.
+   GNAT_PROC, if present is a procedure to call and GNAT_POOL is the
+   storage pool to use.  If not preset, malloc and free will be used.  */
+
+tree
+build_call_alloc_dealloc (gnu_obj, gnu_size, align, gnat_proc, gnat_pool)
+     tree gnu_obj;
+     tree gnu_size;
+     int align;
+     Entity_Id gnat_proc;
+     Entity_Id gnat_pool;
+{
+  tree gnu_align = size_int (align/ BITS_PER_UNIT);
+
+  if (TREE_CODE (gnu_size) != INTEGER_CST
+      && contains_placeholder_p (gnu_size))
+    gnu_size = build (WITH_RECORD_EXPR, sizetype, gnu_size, gnu_obj);
+
+  gnu_size = size_binop (CEIL_DIV_EXPR, gnu_size, size_int (BITS_PER_UNIT));
+
+  if (Present (gnat_proc))
+    {
+      /* The size is the third parameter; the alignment is the same type.  */
+      Entity_Id gnat_size_type
+	= Etype (Next_Formal (Next_Formal (First_Formal (gnat_proc))));
+      tree gnu_size_type = gnat_to_gnu_type (gnat_size_type);
+      tree gnu_proc = gnat_to_gnu_entity (gnat_proc, NULL_TREE, 0);
+      tree gnu_proc_addr = build_unary_op (ADDR_EXPR, NULL_TREE, gnu_proc);
+      tree gnu_pool = gnat_to_gnu_entity (gnat_pool, NULL_TREE, 0);
+      tree gnu_pool_addr = build_unary_op (ADDR_EXPR, NULL_TREE, gnu_pool);
+      tree gnu_args = NULL_TREE;
+      tree gnu_call;
+
+      /* The first arg is always the address of the storage pool; next
+	 comes the address of the object, for a deallocator, then the
+	 size and alignment.  */
+
+      gnu_args
+	= chainon (gnu_args, build_tree_list (NULL_TREE, gnu_pool_addr));
+
+      if (gnu_obj)
+	gnu_args = chainon (gnu_args, build_tree_list (NULL_TREE, gnu_obj));
+
+      gnu_args = chainon (gnu_args,
+			  build_tree_list (NULL_TREE,
+					   convert (gnu_size_type, gnu_size)));
+      gnu_args = chainon (gnu_args,
+			  build_tree_list (NULL_TREE, convert (gnu_size_type,
+							       gnu_align)));
+
+      gnu_call = build (CALL_EXPR, TREE_TYPE (TREE_TYPE (gnu_proc)),
+			gnu_proc_addr, gnu_args, NULL_TREE);
+      TREE_SIDE_EFFECTS (gnu_call) = 1;
+      return gnu_call;
+    }
+  else if (gnu_obj)
+    return build_call_1_expr (free_decl, gnu_obj);
+  else
+    return build_call_1_expr (malloc_decl, gnu_size);
+}
+
 /* Build a GCC tree to correspond to allocating an object of TYPE whose
    initial value is INIT, if INIT is nonzero.  Convert the expression to
-   RESULT_TYPE, which must be some type of pointer.  Return the tree.   */
+   RESULT_TYPE, which must be some type of pointer.  Return the tree.
+   GNAT_PROC and GNAT_POOL optionally give the procedure to call and
+   the storage pool to use.  */
 
 tree
-build_allocator (type, init, result_type)
+build_allocator (type, init, result_type, gnat_proc, gnat_pool)
      tree type;
      tree init;
      tree result_type;
+     Entity_Id gnat_proc;
+     Entity_Id gnat_pool;
 {
   /* Counts number of allocators we were able to do by statically allocating
      memory when at top level.  */
@@ -778,173 +1050,107 @@ build_allocator (type, init, result_type)
   tree ptr_type;
   tree result;
 
-  /* If TYPE is an unconstrained array, it must be the case that INIT
-     is nonzero and either an UNCONSTRAINED_ARRAY_REF or some array-valued
-     expression.  Set SIZE to be the sum of the sizes of the object, its
-     template, and the fat pointer.  Allocate the whole thing and fill in
-     the parts.  */
-  if (TREE_CODE (type) == UNCONSTRAINED_ARRAY_TYPE)
+  /* If RESULT_TYPE is a fat pointer, set SIZE to be the sum of the sizes
+     of the object and its template.  Allocate the whole thing and fill in
+     the parts that are known.  */
+  if (TYPE_FAT_POINTER_P (result_type))
     {
-      tree fat_ptr_type = TYPE_POINTER_TO (type);
-      tree array_type = TREE_TYPE (TREE_TYPE (TYPE_FIELDS (fat_ptr_type)));
+      tree array_type = TREE_TYPE (TREE_TYPE (TYPE_FIELDS (result_type)));
       tree template_type
-	= TREE_TYPE (TREE_TYPE (TREE_CHAIN (TYPE_FIELDS (fat_ptr_type))));
-      tree fat_ptr_size = TYPE_SIZE (fat_ptr_type);
+	= TREE_TYPE (TREE_TYPE (TREE_CHAIN (TYPE_FIELDS (result_type))));
       tree template_size = TYPE_SIZE (template_type);
       tree storage;
-      tree old_template, old_array_ptr, new_template_ptr, new_array_ptr;
-      enum tree_code init_code;
-
-      if (init == 0)
-	gigi_abort (513);
-
-      init_code = TREE_CODE (TREE_TYPE (init));
-      if ((init_code != UNCONSTRAINED_ARRAY_TYPE && init_code != ARRAY_TYPE)
-	  || (init_code == UNCONSTRAINED_ARRAY_TYPE
-	      && type != TREE_TYPE (init)))
-	gigi_abort (514);
-
-      /* If INIT is an unconstrained array, the size is that of the underlying
-	 array; otherwise the size is that of INIT.  */
-      size = (init_code == UNCONSTRAINED_ARRAY_TYPE
-	      ? TYPE_SIZE (array_type) : TYPE_SIZE (TREE_TYPE (init)));
+      tree new_template_ptr, new_array_ptr;
 
       if (TREE_CODE (size) != INTEGER_CST
 	  && contains_placeholder_p (size))
-	size = build (WITH_RECORD_EXPR, sizetype, size,
-		      (init_code == UNCONSTRAINED_ARRAY_TYPE
-		       ? TREE_OPERAND (init, 0) : init));
+	size = build (WITH_RECORD_EXPR, sizetype, size, init);
 
-      size = size_binop (PLUS_EXPR, size,
-			 size_binop (PLUS_EXPR, fat_ptr_size, template_size));
+      size = size_binop (PLUS_EXPR, size, template_size);
 
-      storage = build (CALL_EXPR, build_pointer_type (char_type_node),
-		       build_unary_op (ADDR_EXPR, NULL_TREE, malloc_decl),
-		       build_tree_list (NULL_TREE,
-					size_binop (CEIL_DIV_EXPR,
-						    size,
-						    size_int (BITS_PER_UNIT))),
-		       NULL_TREE);
-      TREE_SIDE_EFFECTS (storage) = 1;
+      storage = build_call_alloc_dealloc (NULL_TREE, size,
+					  MAX (TYPE_ALIGN (array_type),
+					       TYPE_ALIGN (template_type)),
+					  gnat_proc, gnat_pool);
       storage = save_expr (storage);
 
-      /* Skip the fat pointer to get the template, then skip both to get
-	 the memory for the array.  */
-      new_template_ptr
-	= convert (build_pointer_type (template_type),
-		   build (PLUS_EXPR, TREE_TYPE (storage),
-			  storage,
-			  convert (TREE_TYPE (storage),
-				   size_in_bytes (fat_ptr_type))));
+      /* The template is at the start of the allocated storage followed
+	 by the array itself.  */
+      new_template_ptr = convert (build_pointer_type (template_type), storage);
 
       new_array_ptr
-	= convert
-	  (build_pointer_type (array_type),
-	   build (PLUS_EXPR, TREE_TYPE (storage),
-		  storage,
-		  convert (TREE_TYPE (storage),
-			   size_binop (PLUS_EXPR,
-				       size_in_bytes (fat_ptr_type),
-				       size_in_bytes (template_type)))));
+	= convert (build_pointer_type (array_type),
+		   build (PLUS_EXPR, TREE_TYPE (storage), storage,
+			  convert (TREE_TYPE (storage),
+				   size_in_bytes (template_type))));
 
-      /* If this is already an unconstrained array, copy the old template.
-	 Otherwise, build a new one.  There are also different ways of pointing
-	 at the array data.  */
-      if (init_code == UNCONSTRAINED_ARRAY_TYPE)
-	{
-	  tree old_template_ptr
-	    = build_component_ref (TREE_OPERAND (init, 0),
-				   get_identifier ("P_TEMPLATE"), NULL_TREE);
+      /* Copy the template and array (if initializer).  */
+      new_template_ptr
+	= build (COMPOUND_EXPR, build_pointer_type (template_type),
+		 build_binary_op (MODIFY_EXPR, template_type,
+				  build_unary_op (INDIRECT_REF, NULL_TREE,
+						  new_template_ptr),
+				  build_template (template_type, type, init)),
+		 new_template_ptr);
 
-	  old_template
-	    = build_unary_op (INDIRECT_REF, NULL_TREE, old_template_ptr);
-	  old_array_ptr = build_component_ref (TREE_OPERAND (init, 0),
-					       get_identifier ("P_ARRAY"),
-					       NULL_TREE);
-	}
-      else
-	{
-	  old_template
-	    = build_template (template_type, TREE_TYPE (init), init);
-	  old_array_ptr = build_unary_op (ADDR_EXPR, NULL_TREE, init);
-	}
-
-      result = build_unary_op (INDIRECT_REF, fat_ptr_type,
-			       convert (build_pointer_type (fat_ptr_type),
-					storage));
-
-      result = build (COMPOUND_EXPR, fat_ptr_type,
-		      build_binary_op (MODIFY_EXPR, template_type,
-				       build_unary_op (INDIRECT_REF, NULL_TREE,
-						       new_template_ptr),
-				       old_template),
-		      result);
-
-      result = build (COMPOUND_EXPR, fat_ptr_type,
-		      build_binary_op (MODIFY_EXPR, array_type,
-				       build_unary_op (INDIRECT_REF, NULL_TREE,
-						       new_array_ptr),
-				       build_unary_op (INDIRECT_REF, NULL_TREE,
-						       old_array_ptr)),
-		      result);
+      if (init != 0)
+	new_array_ptr
+	  = build (COMPOUND_EXPR, build_pointer_type (array_type),
+		   build_binary_op (MODIFY_EXPR, array_type,
+				    build_unary_op (INDIRECT_REF, NULL_TREE,
+						    new_array_ptr),
+				    build_unary_op (INDIRECT_REF, NULL_TREE,
+						    build_unary_op (ADDR_EXPR,
+								    NULL_TREE,
+								    init))),
+		   new_array_ptr);
 
       return
-	build_binary_op
-	  (MODIFY_EXPR, fat_ptr_type,
-	   result,
-	   build_constructor
-	   (fat_ptr_type,
-	    tree_cons (TYPE_FIELDS (fat_ptr_type), new_array_ptr,
-		       tree_cons (TREE_CHAIN (TYPE_FIELDS (fat_ptr_type)),
-				  new_template_ptr, NULL_TREE))));
+	build_constructor
+	  (result_type,
+	   tree_cons (TYPE_FIELDS (result_type), new_array_ptr,
+		      tree_cons (TREE_CHAIN (TYPE_FIELDS (result_type)),
+				 new_template_ptr, NULL_TREE)));
     }
 
-  /* If TYPE is an discriminated record, see if there is an initializing
-     expression.  If there is, make the object we are allocating that size.
-     Otherwise, allocate an object that is the maximum size of the record.
-     Note that the record must have had default discriminant or this would
-     not be valid Ada.  */
+  /* If we have an initializing expression, see if its size is simpler
+     than the size from the type.  */
+  if (init != 0 && TYPE_SIZE (TREE_TYPE (init)) != 0
+      && (TREE_CODE (TYPE_SIZE (TREE_TYPE (init))) == INTEGER_CST
+	  || (TREE_CODE (size) != INTEGER_CST
+	      && contains_placeholder_p (size))))
+    size = TYPE_SIZE (TREE_TYPE (init));
+
+  /* If the size is still self-referential, reference the initializing
+     expression, if it is present.  If not, this must have been a
+     call to allocate a library-level object, in which case we use
+     the maximum size.  */
   if (TREE_CODE (size) != INTEGER_CST && contains_placeholder_p (size))
     {
-      if (init)
-	{
-	  /* Use INIT's size; we may not have to do anything if it is
-	     a constant.  */
-	  size = TYPE_SIZE (TREE_TYPE (init));
-	  if (TREE_CODE (size) != INTEGER_CST)
-	    size = build (WITH_RECORD_EXPR, sizetype, size, init);
-	}
-      else
+      if (init == 0)
 	size = max_size (size, 1);
+      else
+	size = build (WITH_RECORD_EXPR, sizetype, size, init);
     }
 
   /* If we are at top-level and SIZE is a constant, we can actually
-     allocate an object of TYPE and point to it.  */
+     allocate an object of TYPE and point to it.
+     ??? At some point, we should make an attempt to do this above.  */
   if (global_bindings_p () && TREE_CODE (size) == INTEGER_CST)
     {
       char name[20];
 
       sprintf (name, "__V%d", alloc_var_index++);
-      result = create_var_decl (name, NULL_PTR, type, init, 0, 0, 0, 1);
+      result
+	= create_var_decl (name, NULL_PTR, type, init, size, 0, 0, 0, 0, 1);
       result = build_unary_op (ADDR_EXPR, NULL_TREE, result);
       init = 0;
     }
   else
-    {
-      /* Call the allocator and return an object whose type is a pointer to
-	 TYPE.  Note that we pass the size in bytes and convert to
-	 sizetype.  */
-
-      size = size_binop (CEIL_DIV_EXPR, convert (sizetype, size),
-			 size_int (BITS_PER_UNIT));
-
-      result
-	= build (CALL_EXPR, build_pointer_type (type),
-		 build_unary_op (ADDR_EXPR, NULL_TREE, malloc_decl),
-		 chainon (NULL_TREE, build_tree_list (NULL_TREE, size)),
-		 NULL_TREE);
-      TREE_SIDE_EFFECTS (result) = 1;
-    }
+    result = convert (result_type,
+		      build_call_alloc_dealloc (NULL_TREE, size,
+						TYPE_ALIGN (type),
+						gnat_proc, gnat_pool));
 
   /* If we have an initial value, put the new address into a SAVE_EXPR, assign
      the value, and return the address.  Do this with a COMPOUND_EXPR.  */
@@ -960,18 +1166,9 @@ build_allocator (type, init, result_type)
 		      result);
     }
 
-  /* If the result is not of the proper type, convert it.  But check for
-     converting a pointer to a constrained type into a pointer to
-     an unconstrained type.  In that case, we have to allocate the template
-     from memory as well.  */
+  /* If the result is not of the proper type, convert it.  */
   if (TREE_TYPE (result) != result_type)
-    {
-      if (TYPE_FAT_POINTER_P (result_type)
-	  & TREE_CODE (TREE_TYPE (TREE_TYPE (result))) == ARRAY_TYPE)
-	result = convert_to_unconstrained (result_type, result, 1);
-      else
-	result = convert (result_type, result);
-    }
+    result = convert (result_type, result);
 
   return result;
 }
@@ -1013,27 +1210,3 @@ mark_addressable (tree expr_node)
 	return 1;
     }
 }
-
-#if 0
-----------------------
--- REVISION HISTORY --
-----------------------
-
-----------------------------
-revision 1.72
-date: Mon Jul 11 09:59:10 1994;  author: kenner
-(build_unary_op): Allow TYPE to be error_mark_node.  [Partial fix
-for 0701-001].
-----------------------------
-revision 1.73
-date: Fri Jul 15 18:36:02 1994;  author: schonber
-(build_unary_op):for modular types, same treatment as in build_binary_op
-[fixes crash on abs, reported by Keith].
-----------------------------
-revision 1.74
-date: Mon Jul 18 16:37:10 1994;  author: banner
-(build_allocator): change p_array to P_ARRAY and p_template to P_TEMPLATE.
- change requested by Paul for interaction with gdb.
-----------------------------
-** New changes after this line and before endif. **
-#endif
